/*
 *                OMEGA IRC SECURITY SERVICES
 *                  (C) 2009, Jeremy Johnston <jeremy at ssnet dot ca>
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *    $Id: m_splitserv.c 1698 2010-01-07 02:27:06Z jer $
 */

/*
 * This module provides similiar functionality as SplitServ
 * found in Sentinel - IRC Statistical and Operator Services.
 */

#include "stdinc.h"
#include "server.h"

#define NAME    "m_splitserv"
#define AUTHOR  "Jer"
#define VERSION "$Id: m_splitserv.c 1698 2010-01-07 02:27:06Z jer $"

#define SPLITSERV_NICK  "SplitServ"
#define SPLITSERV_USER  "splitserv"
#define SPLITSERV_HOST  "security."
#define SPLITSERV_CHAN  "#splits"

/* If you would like SplitServ to track splits
 * across reboots change this to #define
 */
#undef  SPLITSERV_SAVE

/* Allow expiring of splits */
#define SPLITSERV_EXPIRE
#define SPLITSERV_EXPIRE_TIMEOUT    "2d"

/* STOP STOP STOP STOP STOP STOP
 * Do not edit below this line unless you know what
 * you are doing!!!
 */

#define TIMEBUF		100

/* These items are for the timediff() function, used to process time
** quantities (3 days, 45 minutes, 21 seconds)
*/
#define T_DAY       "day"
#define T_DAYS      "days"
#define T_HOUR      "hour"
#define T_HOURS     "hours"
#define T_MINUTE    "minute"
#define T_MINUTES   "minutes"
#define T_SECOND    "second"
#define T_SECONDS   "seconds"

#define BUFSIZE     512

/* Init/Deinit */
int     splitserv_init ();
void    splitserv_deinit ();

/* Commands */
static void splitserv_cmd_split     (User *, char **);
static void splitserv_cmd_forget    (User *, char **);
static void splitserv_cmd_add       (User *, char **);

static void splitserv_cmd_privmsg   (User *, int, char **);

static void splitserv_show_splits   (char *);
static void splitserv_show_map      (char *);
static struct SplitServer *splitserv_addsplit      (char *, char *, int);
static void splitserv_forgetsplit   (char *);

struct SplitServer  *splitserv_findsplit    (char *);

/* Event Handlers */
void    splitserv_servconnect   (int, Link *);
void    splitserv_servexit      (int, Link *);
void    splitserv_expiresplits  (void);

/* Split Tracking */
struct SplitServer {
    char    servname[63];
    char    uplink[63];

    time_t  splittime;
};

dlink_list  splits;     /* Linked list of split servers (struct *SplitServer) */

/* Tracking splits across reboots */
static char     splitserv_db[MAXPATH];

void    splitserv_savedb    (void);
void    splitserv_loaddb    (void);

MODHEADER (NAME, VERSION, AUTHOR, MAKE_ABI(0,5,2), splitserv_init, splitserv_deinit);

User *s_SplitServ = NULL;

int splitserv_init ()
{
    Channel     *chan;
    int         ch;

    splits.head = splits.tail = NULL;

    snprintf (splitserv_db, sizeof(splitserv_db), "%s/splitsdb", datadir);

    if (s_SplitServ != NULL)
        return MOD_STOP;

    if (!(s_SplitServ = NewClient (SPLITSERV_NICK, SPLITSERV_USER, SPLITSERV_HOST, "Split Tracking Services")))
        return MOD_STOP;

    if (!(chan = find_channel (SPLITSERV_CHAN)))
    {
        chan            = new_chan (SPLITSERV_CHAN);
        chan->channelts = time(NULL);
    }

    AddToChannelU (s_SplitServ, chan);

    splitserv_loaddb ();

    AddEvent ("SERVERCONN", splitserv_servconnect);
    AddEvent ("SERVEREXIT", splitserv_servexit);

    ch  = AddCmd (s_SplitServ, "SPLIT", ACC_FLAG_OPER, splitserv_cmd_split);
    ch  = AddCmd (s_SplitServ, "FORGET", ACC_FLAG_OPER, splitserv_cmd_forget);
    ch  = AddCmd (s_SplitServ, "ADD", ACC_FLAG_OPER, splitserv_cmd_add);

    AddUserCmd ("PRIVMSG", splitserv_cmd_privmsg);

    AddTimedEvent ("Split expire", splitserv_expiresplits, 5);
    AddTimedEvent ("Split db save", splitserv_savedb, 900);

    return MOD_CONT;
}

void splitserv_deinit ()
{
    struct SplitServer  *sp;
    dlink_node          *dl, *tdl;

    DelEvent ("SERVERCONN", splitserv_servconnect);
    DelEvent ("SERVEREXIT", splitserv_servexit);

    DelTimedEvent ("Split expire");
    DelTimedEvent ("Split db save");

    splitserv_savedb ();

    if (s_SplitServ) {
        DelCmd (s_SplitServ, "SPLIT", splitserv_cmd_split);
        DelCmd (s_SplitServ, "FORGET", splitserv_cmd_forget);
        DelCmd (s_SplitServ, "ADD", splitserv_cmd_add);

        DelUserCmd ("PRIVMSG", splitserv_cmd_privmsg);

        DelClient (s_SplitServ->nick);
    }

    /* Be sure to release all split entries */
    DLINK_FOREACH_SAFE (dl, tdl, splits.head)
    {
        sp  = (struct SplitServer *) dl->data;

        dlink_delete (dl, &splits);
        dlink_free (dl);

        free (sp);
    }

    return;
}

void splitserv_loaddb (void)
{
#ifdef SPLITSERV_SAVE
#ifdef HAVE_SQLITE
    struct sqlite3      *db;
    struct SplitServer  *sp;

    dlink_node          *dl;

    char                **data, *errmsg;
    char                *server, *uplink;
    int                 i, j, row_count, col_count, pos, splittime;

    alog (DEBUG, "Attempting to load %s", splitserv_db);

    if (sqlite3_open (splitserv_db, &db))
    {
        alog (LOG_ERROR, "Unable to open splits db");
        return;
    }

    if ((i = sqlite3_get_table (db, "SELECT * FROM splits", &data, &row_count, &col_count, &errmsg)))
    {
        switch (i)
        {
            case SQLITE_BUSY:
                break;

            default:
                alog (LOG_ERROR, "Problem with splitsdb: %s", errmsg);
                break;
        }

        sqlite3_close (db);
        return;
    }

    if (row_count == 0)
        return;

    pos = col_count;

    for (i = 0; i < row_count; i++)
    {
        for (j = 0; j < col_count; j++)
        {
            switch (j)
            {
                case 0:
                    server  = data[pos++];
                    break;
                case 1:
                    uplink  = data[pos++];
                    break;
                case 2:
                    splittime   = atoi(data[pos++]);
                    break;
            }
        }

        sp  = splitserv_addsplit (server, uplink, splittime);

    }

    if (data) {
        fprintf (stderr, "freeing %p\n", data);
        //free (data);
    }

    if (db)
        sqlite3_close (db);

#endif /* HAVE_SQLITE */

    alog (LOG_INFO, "Split saving is disabled. Missing SQLite3 support");
#endif /* SPLITSERV_SAVE */
}

void splitserv_savedb (void)
{
#ifdef SPLITSERV_SAVE
#ifdef HAVE_SQLITE
    struct sqlite3      *db;
    struct SplitServer  *sp;
    dlink_node          *dl;

    int                 i;
    char                *errmsg;
    static char         buf[BUFSIZE * 4];

    if (sqlite3_open (splitserv_db, &db))
    {
        alog (LOG_ERROR, "Unable to open splits db");
        return;
    }

    if ((i = sqlite3_exec (db, "DELETE FROM splits", NULL, NULL, &errmsg)))
    {
        alog (LOG_ERROR, "Problem with splitsdb: %s", errmsg);
        ircd_wallop (s_SplitServ->nick, "Error during split database save");

        sqlite3_close (db);
        return;
    }

    DLINK_FOREACH (dl, splits.head)
    {
        sp  = (struct SplitServer *) dl->data;

        snprintf (buf, sizeof(buf), "INSERT INTO splits VALUES ('%s', '%s', %d)", sp->servname, sp->uplink, (int) sp->splittime);

        if ((i = sqlite3_exec (db, buf, NULL, NULL, &errmsg)))
        {
            alog (LOG_ERROR, "Problem with splitdb: %s", errmsg);
            ircd_wallop (s_SplitServ->nick, "Error during split database save");

            sqlite3_close (db);
            return;
        }
    }

    sqlite3_close (db);
    return;

#endif /* HAVE_SQLITE */
#endif /* SPLITSERV_SAVE */
}

void splitserv_servconnect (int unused, Link *serv)
{
    dlink_node          *dl, *tdl;
    struct SplitServer  *sp;

    if ((strcasecmp (serv->name, CfgSettings.servername)) == 0) /* Ignore ourself */
        return;

    DLINK_FOREACH_SAFE (dl, tdl, splits.head)
    {
        sp  = (struct SplitServer *) dl->data;

        if (((strcasecmp (sp->servname, serv->name)) == 0) && ((strcasecmp (sp->uplink, serv->uplink->name)) == 0))
        {
            sendto_channel (s_SplitServ, find_channel(SPLITSERV_CHAN), "END OF SPLIT: %s to %s: %s", sp->servname, sp->uplink, timediff(time(NULL) - sp->splittime));

            free (sp);
            dlink_delete (dl, &splits);
            dlink_free (dl);
            return;
        }
    }

    sendto_channel (s_SplitServ, find_channel(SPLITSERV_CHAN), "CONNECT: %s to %s", serv->name, (serv->uplink)? serv->uplink->name : "unknown");
}

void splitserv_servexit (int unused, Link *serv)
{
    dlink_node          *dl;
    struct SplitServer  *sp;

    if ((strcasecmp (serv->name, CfgSettings.servername)) == 0) /* Ignore ourself */
        return;

    sp  = (struct SplitServer *) malloc (sizeof(struct SplitServer));
    dl  = dlink_create ();

    strlcpy (sp->servname, serv->name, sizeof(sp->servname));

    if (serv->uplink)
        strlcpy (sp->uplink, serv->uplink->name, sizeof(sp->uplink));
    else
        strlcpy (sp->uplink, "null", sizeof(sp->uplink));

    sp->splittime   = time(NULL);

    sendto_channel (s_SplitServ, find_channel(SPLITSERV_CHAN), "SPLIT: %s from %s", sp->servname, sp->uplink);

    dlink_add_tail (sp, dl, &splits);
}

void splitserv_expiresplits (void)
{
#ifdef SPLITSERV_EXPIRE
    struct SplitServer  *sp;
    dlink_node          *dl, *tdl;
    int expire_time;

    expire_time = 172800; /* 2 days */

    DLINK_FOREACH_SAFE (dl, tdl, splits.head)
    {
        sp  = (struct SplitServer *) dl->data;

        if ((time(NULL) - sp->splittime) > expire_time)
        {
            sendto_channel (s_SplitServ, find_channel(SPLITSERV_CHAN), "SPLIT EXPIRED AFTER %s: %s", timediff(time(NULL) - sp->splittime), sp->servname);

            dlink_delete (dl, &splits);
            dlink_free (dl);

            free (sp);
        }
    }
#endif /* SPLITSERV_EXPIRE */

    return;
}

static void splitserv_cmd_privmsg (User *source, int argc, char **argv)
{
    char    *av[5];
    int     ac;

    if (!source)
        return;

    if ((*argv[1] != '#') || (*argv[2] != '!'))
        return;

    if ((strcasecmp (SPLITSERV_CHAN, argv[1])) != 0)
        return;

    ac  = generictoken (' ', 5, argv[2], av);

    av[0]++;    // Skip the fantasy character

    if ((strcasecmp (av[0], "SPLIT")) == 0)
    {
        splitserv_show_splits (argv[1]);
        return;
    }

    if ((strcasecmp (av[0], "ADD")) == 0)
    {
        if (ac >= 3)
            splitserv_addsplit (av[1], av[2], 0);
        return;
    }

    if ((strcasecmp (av[0], "FORGET")) == 0)
    {
        if (ac >= 2)
            splitserv_forgetsplit (av[1]);
        return;
    }

    if ((strcasecmp (av[0], "MAP")) == 0)
    {
        splitserv_show_map (argv[1]);
        return;
    }
}

static void splitserv_cmd_split (User *user, char **av)
{
    splitserv_show_splits (user->nick);
}

static void splitserv_cmd_forget (User *user, char **av)
{
    if (!av[0])
        return;
    splitserv_forgetsplit (av[0]);
}

static void splitserv_cmd_add (User *user, char **av)
{
    if (!av[0] || !av[1])
        return;

    splitserv_addsplit (av[0], av[1], 9);
}

static void splitserv_show_splits (char *target)
{
    dlink_node          *dl;
    struct SplitServer  *sp;
    int                 cnt = 0;

    Channel      *c = NULL;
    User         *u = NULL;

    if (*target == '#')
        c = find_channel (target);
    else
        u = find_user (target);


    if (!c && !u)
        return;

    DLINK_FOREACH (dl, splits.head)
    {
        sp = (struct SplitServer *) dl->data;

        cnt++;

        if (c)
            sendto_channel (s_SplitServ, c, "SPLIT: %s from %s: %s", sp->servname, sp->uplink, timediff(time(NULL) - sp->splittime));
        else
            sendto_one (s_SplitServ, u, "SPLIT: %s from %s: %s", sp->servname, sp->uplink, timediff(time(NULL) - sp->splittime));
    }

    if (cnt == 0) {
        if (c)
            sendto_channel (s_SplitServ, c, "NO SPLITS");
        else
            sendto_one (s_SplitServ, u, "NO SPLITS");
    }

    return;
}

static void splitserv_show_map (char *target)
{
    return;
}

static struct SplitServer *splitserv_addsplit (char *server, char *uplink, int ts)
{
    struct SplitServer  *sp;
    Link                *serv;
    dlink_node          *dl;

    if ((sp = splitserv_findsplit (server)) != NULL)
        return NULL;

    if ((serv = find_serv (server)) != NULL)
        return NULL;

    sp  = (struct SplitServer *) malloc (sizeof (struct SplitServer));
    dl  = dlink_create ();

    strlcpy (sp->servname, server, sizeof(sp->servname));
    strlcpy (sp->uplink, uplink, sizeof(sp->uplink));

    if (ts == 0)
        sp->splittime   = time(NULL);
    else
        sp->splittime   = ts;

    dlink_add_tail (sp, dl, &splits);

    sendto_channel (s_SplitServ, find_channel(SPLITSERV_CHAN), "SPLIT: %s from %s", sp->servname, sp->uplink);

    return sp;
}

static void splitserv_forgetsplit (char *server)
{
    struct SplitServer  *sp;
    dlink_node          *dl, *tdl;

    if ((sp = splitserv_findsplit (server)) == NULL)
        return;

    sendto_channel (s_SplitServ, find_channel(SPLITSERV_CHAN), "FORGOT SPLIT: %s", sp->servname);

    dl = dlink_find_delete (sp, &splits);

    dlink_free (dl);
    free (sp);

}

struct SplitServer *splitserv_findsplit (char *server)
{
    struct SplitServer  *sp;
    dlink_node          *dl;

    DLINK_FOREACH (dl, splits.head)
    {
        sp  = (struct SplitServer *) dl->data;

        if ((strcasecmp (sp->servname, server)) == 0)
            return sp;
    }

    return NULL;
}

/* print a period of seconds into human readable format */
char *timediff(time_t u)
{
  static char buf[TIMEBUF];

  if (u > 86400)
  {
    time_t days = u / 86400;

    snprintf(buf, TIMEBUF, "%ld %s, %02ld:%02ld",
             (long)days, (days == 1) ? T_DAY : T_DAYS,
             (long)((u / 3600) % 24), (long)((u / 60) % 60));
  }
  else if (u > 3600)
  {
    time_t hours = u / 3600;
    time_t minutes = (u / 60) % 60;

    snprintf(buf, TIMEBUF, "%ld %s, %ld %s",
             (long)hours, (hours == 1) ? T_HOUR : T_HOURS, (long)minutes,
             (minutes == 1) ? T_MINUTE : T_MINUTES);
  }
  else
  {
    time_t minutes = u / 60;
    time_t seconds = u % 60;

    snprintf(buf, TIMEBUF,
             "%ld %s, %ld %s",
             (long)minutes, (minutes == 1) ? T_MINUTE : T_MINUTES,
             (long)seconds, (seconds == 1) ? T_SECOND : T_SECONDS);
  }

  return buf;
}
