/*
 *         	  OMEGA IRC SECURITY SERVICES
 * 	      	    (C) 2008 Omega Dev Team
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *    $Id: access.c 1725 2010-01-13 06:49:41Z twitch $
 */


#include "stdinc.h"
#include "server.h"


/*******************************************/
/**
 * find_access() - Find the FIRST matching access
 * access list entry for the given user and return
 * the access list entry.
 *
 * @param u - Pointer to the user struct
 *            that we are checking with.
 *
 * @return Access* - return a pointer to the
 *                   access entry list that the user
 *                   that the user belongs to, or
 *                   NULL if no match is found.
 *
 */


Access* find_access(User *u)
{
	dlink_node *dl,*hdl;
	char *host;
	Access *a;
	char host_pattern[150];
	char ip_pattern[150];
	char nick_pattern[250];
	
    snprintf(host_pattern,sizeof(host_pattern),"%s@%s",(u->user)? u->user : "*",u->host);
	snprintf(ip_pattern,sizeof(ip_pattern),"%s@%s",(u->user)? u->user : "*",u->ip);
	
	DLINK_FOREACH(dl,accesslist.head)
	{
		a = dl->data;

		DLINK_FOREACH(hdl,a->hosts.head)
		{
		 	if (!(host = hdl->data))
					return NULL;
			if (strchr(host,'!'))
			{
					memset(nick_pattern,0,sizeof(nick_pattern));
					snprintf(nick_pattern,sizeof(nick_pattern),"%s!%s",u->nick,host_pattern);
				
					if (match_wild(host,nick_pattern,0))
						return a;
						
					memset(nick_pattern,0,sizeof(nick_pattern));
					snprintf(nick_pattern,sizeof(nick_pattern),"%s!%s",u->nick,ip_pattern);
					
					if (match_wild(host,nick_pattern,0))
						return a;
						
			} else {
				if (match_wild(host,host_pattern,0))
					return a;
				if (match_wild(host,ip_pattern,0))
					return a;
			}
			continue;
		}
	}
	return NULL;
}


/*******************************************/
/**
 * Check a users access against the specified level.
 *
 * @param u A pointer to the user structure which
 *          we are checking against.
 * @param level The level to check for.
 *
 * @return Return true if the access match, or return
 *         0 if we fail for any other reason.
 */


int check_access(User* u, char *level)
{

	if (!u)
		return 0;

	if (!level)
		return 1;

	//user has no access
	if (!u->access)
		return 0;

	if (strchr(u->access,*level))
		return 1;

	return 0;


}

/*******************************************/
/**
 * setAccess() - finds then sets or removes
 * access for a user, this will set the first
 * matching access level on a user, as provided '
 * by find_access().
 *
 *  @param u - The user struct we are working with
 *
 *
 *  @return int - Returns true if access has been set
 *                or false if no record was found.
 */


int setAccess(User* u)
{

	Access *a;

	if (!u)
		return 0;

	memset(u->access,0,sizeof(u->access));
	strlcpy(u->access,ACC_FLAG_USER,sizeof(u->access));
	
    if ((!u->oper) && (CfgSettings.operonly))
        return 0;
		
	//standard oper access
	if (u->oper)
		strncat (u->access, ACC_FLAG_OPER, sizeof(u->access));	
		
    if (!(a = find_access (u))) 
        return 0;
		
    strncat (u->access, a->flags, sizeof(u->access));
	return 1;
}

/*******************************************/

Access *access_exists(char *name)
{
		dlink_node *dl;
		Access *a;
		
		DLINK_FOREACH(dl,accesslist.head)
		{
				a = dl->data;
				if (strcasecmp(a->name,name)==0)
					return a;
		
		}
		return NULL;
}

/*******************************************/

Access *new_access(char *name)
{
		Access *a;
		dlink_node *dl;
		
		if (a = access_exists(name))
				return a;

        alog (LOG_DEBUG, "new_access(): Creating new access for %s", name);
				
		dl = dlink_create();
		if (!(a  = (Access *) malloc(sizeof(Access))))
				return NULL;
		
		memset(a->name, '\0',sizeof(a->name));
		memset(a->flags,'\0',sizeof(a->flags));
	    memset(a->cname,'\0',sizeof(a->cname));
		
		strlcpy(a->name,name,sizeof(a->name));
		strlcpy(a->flags,	ACC_USER,	sizeof(a->flags));
		
		a->hosts.head	= a->hosts.tail	= NULL;
		
		dlink_add_tail (a, dl, &accesslist);
		return a;
}


/*******************************************/

int add_access_host(Access* a, char *host) 
{
	dlink_node *dl;
	char *tmp_host;
	char *ac_host; 
	
	if (!a)
			return 0;
	if (!host)
			return -1;

    alog (LOG_DEBUG, "add_access_host(): Adding host %s to %s", a->name, host);
			
	DLINK_FOREACH(dl,a->hosts.head)
	{
			ac_host = dl->data;
			
			if (strcasecmp(ac_host,host)==0)
					return 0;
	}
	dl	= dlink_create();
	tmp_host = strdup(host);
	dlink_add_tail (tmp_host, dl, &a->hosts);
	return 1;
}

/**
 * Use pointer logic in the future
 */

int set_access_flag(Access *a, char *flags)
{
	AccessFlag *af;
	int i = 0;
	dlink_node *dl;
    char a_flags[128];
	int len = 0;
	
	if (!*a->flags) {
		memset(a->flags,0,sizeof(a->flags));
		strlcat(a->flags, ACC_USER, sizeof(a->flags));
	}
	
	memset(a_flags,0,sizeof(a_flags));
	

    alog (LOG_DEBUG, "Processing access flags for %s", a->name);
    alog (LOG_DEBUG, "set_access_flag(): requested flags: %s", flags);
	//handle name value access flags
	if (af = find_access_flag(flags)) {
		strlcat(a_flags,af->flag,sizeof(a_flags));
		
        alog (LOG_DEBUG, "Adding %s flag to %s", af->flag, a->name);
        alog (LOG_DEBUG, "a_flags set to %s", a_flags);
		
		len = strlen(a->flags);
		
		for (i = 0; a_flags[i] != '\0'; i++)
		{
			alog (LOG_DEBUG, "set_access_flags(): a_flags[%d]: %c", i, a_flags[i]);
			if (!strchr(a->flags, a_flags[i])) {
				if (len < sizeof(a->flags)) {
					a->flags[len] = a_flags[i]; //this would overrite flags already set
					len++;
				} else {
					break;
				}
			} 
		}
	//handle flag form access.
	} else {
		len = strlen(a->flags);
		for (i = 0; flags[i]; i++) {
            alog (LOG_DEBUG, "set_access_flags(): flags[%d]: %c", i, flags[i]);
			if ((af = find_flag_byvalue(flags[i]))) {
                alog (LOG_DEBUG, "Adding %s flag to %s", af->flag, a->name);
				
				if (!strchr(a->flags, flags[i])) {
					if (len < sizeof(a->flags)) {
						a->flags[len] = flags[i]; //this would overrite flags already set
						len++;
					} else {
						break;
					}
				}
			}
		}
	}
    alog (LOG_DEBUG, "Set access flags on %s to %s", a->name, a->flags);
	a->flags[len+1] = '\0';
}
/*************************************************/
/**
 *   Destroy all entries within the access list.
 *
 *   @note This destroyes all access list entries
 *         with that said, it shouldnt be called
 *         in any other circumstances besides rehash
 *         and shutdown, as it leaves us with no entries.
 *
 *
 */


int DestAccList()
{
	dlink_node *dl;
	dlink_node *tdl;

	dlink_node *hdl,*thdl; //host dlink list
	char *host;

	Access *a;

	DLINK_FOREACH_SAFE(dl,tdl,accesslist.head)
	{

		a = dl->data;

		DLINK_FOREACH_SAFE(hdl,thdl,a->hosts.head)
		{
			host = hdl->data;

			dlink_find_delete(host,&a->hosts);
			dlink_free(hdl);
			
			if (host)
				free(host);
		}


		dlink_find_delete(a,&accesslist);
		dlink_free(dl);
		free(a);

	}
	return 0;

}

/*************************************************/
/**
 *   refresh_access() - Crawls our user list and checks
 *   access on each entries host.
 *
 *   @note This method can be slower on larger nets and
 *   is designed only to update access on special instances
 *   such as rehash.
 *
 *  @XXX Crawling this big of a list is rediculous... i think we should keep a master list somewere
 * -Twitch
 */

void refresh_access(void)
{
    User* u = NULL;
    Access* a = NULL;

    dlink_node *udl = NULL;


     DLINK_FOREACH(udl,userlist.head)
     {
            u = udl->data;

		/**
		 * if there not an OPER or they are a SERVICE
         * they have no access. so skip that entry :)
         * this should give us back some speed and
         * proformance here. Esp on networks with more
         * Services and Users then IRCops (Which should
         * be all of them imo)
         * - Twitch
		 */

		if (((!u->oper) && (CfgSettings.operonly)) || (u->service))
			continue;

		setAccess(u);
     }

    return;
}
/*************************************************/

int add_access_flag (char *name,char *flag)
{
		dlink_node *dl;
		AccessFlag *af;

        alog (LOG_DEBUG, "add_access_flag(): %s / %s", name, flag);
		
		if ((af = find_access_flag(name)))
				return -1;
		if ((af = find_access_flag(flag)))
				return -1;
				
		af = (AccessFlag*) malloc(sizeof(AccessFlag));
		if (!af)
				return 0;
		
		dl = dlink_create();
		
		strlcpy(af->name,name,sizeof(af->name));
		strlcpy(af->flag,flag,sizeof(af->flag));
		dlink_add_tail (af, dl, &access_flags);
		return 1;
}

/*************************************************/

void init_access_flags(void)
{
		add_access_flag("REHASH",ACC_FLAG_REHASH);
		add_access_flag("ACCESS",ACC_FLAG_ACCESS);
		add_access_flag("LIST",ACC_FLAG_LIST);
		add_access_flag("JOIN",ACC_FLAG_JOIN);
		add_access_flag("SET",ACC_FLAG_SET);
		add_access_flag("DIE",ACC_FLAG_DIE);
		add_access_flag("DEBUG",ACC_FLAG_DEBUG);
		add_access_flag("MODULE",ACC_FLAG_MODULE);
		add_access_flag("OPER", ACC_FLAG_OPER);

		add_access_flag("ROOT",ACC_ROOT);
		add_access_flag("SPOPER",ACC_STAFF);
		add_access_flag("STAFF",ACC_STAFF);
		add_access_flag("ADMIN",ACC_ADMN);
		add_access_flag("OPER", ACC_OPER);
		
}

/***********************************************/
/**
 * Return an access flag, both by name or by value
 * 
 * @param name string Name or value of the flag we are looking for
 * @return AccessFlag* pointer to the flag structure.
 *
 */
AccessFlag *find_access_flag(char *name)
{
		dlink_node *dl;
		AccessFlag *af;
		
		if (!name)
			return NULL;
 
        alog (LOG_DEBUG, "find_access_flag(): searching for %s", name);

		DLINK_FOREACH(dl,access_flags.head)
		{
				af = dl->data;

                alog (LOG_DEBUG2, "find_access_flag(): (%s == %s) (%s == %s)", af->name, name, af->flag, name);
				
				if (strcasecmp(name,af->name)==0)
					return af;
				//if (strcmp(name,af->flag)==0)
				//	return af;
		}
		return NULL;
}


AccessFlag *find_flag_byvalue(int flag)
{

		dlink_node *dl;
		AccessFlag *af;
		
		DLINK_FOREACH(dl,access_flags.head)
		{
				af = dl->data;
				
				if (strlen(af->flag) > 1)
					continue;
					
				if (*af->flag == flag)
					return af;
		}
	return NULL;
}

/***********************************************/

int del_acces_flag(char *name)
{
		dlink_node *dl;
		AccessFlag *af;
		
		if (!name)
			return 0;
			
		if ((af = find_access_flag(name)))
		{
			dl  = dlink_find_delete (af, &userlist);
			dlink_free (dl);
			free(af);
			return 1;
		}
		
		return 0;
}

/* EOF */

