/*
 *         	  OMEGA IRC SECURITY SERVICES
 * 	      	    (C) 2008 Omega Dev Team
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *   Holds Client command hash functions.  This is for client commands so
 *   clients do not need to wade through to find there own commands.
 *
 *    $Id: cmds.c 1667 2010-01-04 01:40:42Z twitch $
 */



#include "stdinc.h"
#include "server.h"



const unsigned int prime_table_length = sizeof(primes)/sizeof(primes[0]);
const float max_load_factor = 0.65;

extern Module* currentModule;


int hc = 0;

void HelpNull(User*);


/******************************************************/
/**
 * Add a command to our clients command hash.
 *
 * @note There is a bit of presumptions we make when issueing this
 * function, for one we assume that u is always our client,
 * since client structuresare the same as user structures except
 * they have been flagged as our clients.
 *
 * @param u Pointer to our clients structure
 * @param name The name of the command we are using, this is imporntant
 *        becuase the command will be issued and looked up via its name.
 * @param access Access level required to issue the command.
 * @param fnct A pointer to the command's function.
 *
 * @return True/False - Returns true if everything goes as planned and false if not.
 *
 */


int AddCmd(User* u,char* name,char *access,void (*fnct)(User*,char**))
{

    dlink_node *dl = dlink_create();
	CmdHash *c = (CmdHash*) malloc(sizeof(CmdHash));

	int index = HASH(name,1024);

	c->name = strdup(name);
	
	memset(c->access,0,sizeof(c->access));	
	
	if (access)
		strlcpy(c->access,access,sizeof(c->access));
		
	c->fnct = fnct;
	c->hascmd = 1;
	c->hashlp = 0;
	c->hashlpext = 0;
    alog(DEBUG3,"Adding CLIENT command %s (%d) (Location: %p)",name,index,fnct);
    dlink_add_tail(c,dl,&u->cmd_list[index]);
	return index;


}


/*****************************************************/

int AddHelp(User* u, void (*helpcmd)(User*))
{

    HelpEntry *ch,*ch2;
    dlink_node *dl;


    alog(DEBUG3,"Adding help (Location: %p)",helpcmd);

    if (!u)
	return -1;

   dl = dlink_create();

   ch = (HelpEntry*) malloc(sizeof(HelpEntry));

   ch->routine = helpcmd;

   dlink_add_tail(ch,dl,&u->cmd_help);

   return 1;


}

int AddHelpExt(int index,User* u, void (*helpcmd)(User*))
{

    HelpEntry *hc;
    dlink_node *dl;

    if (!index)
        return -1;

    dl = dlink_create();

    hc = (HelpEntry*) malloc(sizeof(HelpEntry));

    hc->routine = helpcmd;

    dlink_add_tail(hc,dl,&u->help_ext_list[index]);

    alog(DEBUG3,"Adding EXTENDED HELP For command index (%d) (Location: %p)",index,helpcmd);

    return 0;
}


int DelHelp(User *u, void (*helpcmd)(User*))
{

	HelpEntry* he;
	dlink_node *dl;

	DLINK_FOREACH(dl,u->cmd_help.head)
	{
		he = dl->data;

		if (he->routine != helpcmd)
			continue;

		dl = dlink_find_delete(he,&u->cmd_help);
		dlink_free(dl);

		if (he)
			free(he);

		break;
	}
	return;
}

/******************************************************/


void DelCmd(User* u, char* cmd, void (*fnct)(User*,char**))
{
    dlink_node *dl,*tdl;
    CmdHash *ch;


    int index = HASH(cmd,1024);

    DLINK_FOREACH_SAFE(dl,tdl,u->cmd_list[index].head)
    {
            ch = dl->data;

        if (ch->fnct != fnct)
                continue;


           dl = dlink_find_delete(ch,&u->cmd_list[index]);
           dlink_free(dl);

           MyFree(ch);

		   break;
    }

    return;
}


/*****************************************************/
/**
 * int do_cmd() - Handles a command to a client -
 *                do all the santatity checks here.
 *
 *  @param cmd         = The command we are actioning on.
 *  @param us          = A pointer to OUR client struct
 *  @param u           = A pointer to the SRC of the command
 *  @param ac          = The argument count being passed into the command
 *                       handler - used to see if av is null :)
 *  @param av          = The PARAMATERS to the command - the original
 *                       command.
 *  @return int
 *             0 = Success
 *             1 = No cmd for it was found.
 *             2 = Access Denied.
 *             3 = Fail something was passed wrong
 * @TODO 
 *      - Write this so that it can make more sense in the future
 * 		- Also find a better way to handle user commands, i fear that this isnt 
 *        the best method of doing this... meh.
 */



int do_cmd(char *cmd, User* us, User *u, int ac, char *av)
{
 	   CmdHash* c;
	   HelpEntry *he;
       dlink_node *dl;

	   int index, n, found = 0;
	   char* params[512];
	   
 	   if (!us || !u)
 	     	return 0;

       if (strcasecmp(cmd,"HELP")==0)
	   {
           	 if (ac == 2)   {
				  index = HASH(av,1024);
				  if ((dl = us->cmd_list[index].head)) { //check to see if there is anything at index
								if (!(c = dl->data)) {
									sendto_one (us, u, "No such command");
									return 1;
								}
				  }
				  //we really dont have to respect operonly here cause
				  //our access system takes care of that.
				  if (!check_access(u,c->access))  {	//check access 
						sendto_one(us,u,"You do not have the correct permissions to view the help for this command.");
						return 2;
				  }
				  DLINK_FOREACH(dl,us->help_ext_list[index].head)
				  {
						he = dl->data;
						he->routine(u);
						found = 1;
				  }
				  if (!found)
					sendto_one(us,u,"There is no help available for that command.");
				return 0;
            }
            HELP_FOREACH(us,u);
            return 0;
	   }
	   
	   index = HASH(cmd,1024);
	   
	   /*
	    * Call the commands associated with the index 
		* this is called from head to tail
		*/
		n = generictoken(' ',512,av,params);

		if (!*params[0])
			n = 0;
			
	   DLINK_FOREACH(dl,us->cmd_list[index].head)  {
         	c = dl->data;
          	if (strcasecmp(c->name,cmd)==0) {
				  //we really dont have to respect operonly here cause
				  //our access system takes care of that.
                 if (!check_access(u, c->access)) {
                     sendto_one(us,u,"Permission Denied.");
			  		 return 1;
				 }
                 if (!c->fnct)
                        	continue; //no command pointer found bail
                 
                 c->fnct(u,params);
                 found = 1;
          	}
	   }
       return (found)? 0 : 1;
}



/*****************************************************/

int HELP_FOREACH(User* us, User *u)
{
    HelpEntry* he;
    dlink_node *dl;

	DLINK_FOREACH(dl,us->cmd_help.head)
	{
		he = dl->data;
		he->routine(u);
	}
    return 0;
}



/*****************************************************/


void ACCESS_DENIED(User* s,User *u) { return; }


char *CmdErr(int err)
{
	const char *cmd_err_codes[] = {
		"Command success, no error.",
		"Invalid command - No such command found.",
		"Permission Denied.",
		"Command Error, Invalid paramaters.",
		"Command Error, Invalid index.",
		"Command Error, HASH does not exsist."
	};

	return (char*) cmd_err_codes[err];
}

/*****************************************************/


void HelpNull(User *u)
{
        return;
}
